'' =================================================================================================
''
''   File....... jm_hc8_menorah__2016-12-18.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2016 JonnyMac
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 18 DEC 2016
''
'' =================================================================================================


con

  ' Important: You must set this costant to match your PCB!

  PCB_REV = "E"


con { timing }

  _clkmode = xtal1 + pll16x                                      
  _xinfreq = 5_000_000                                          ' use 5MHz crystal
                                                                 
  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  EFX_BAUD =  38_400
  TRM_BAUD = 115_200

  
con { io pins }

  RX1     = 31  { I }                                           ' serial / programming
  TX1     = 30  { O }                                            
                                                                 
  SDA1    = 29  { I/O }                                         ' i2c / eeprom
  SCL1    = 28  { I/O }                                          
                                                                 
  OUT_EN  = 27  { O }   { XP0 on Rev F pcb }                    ' outputs enable pin (low)     
                                                                 
  XP0     = 27  { I/O } { OUT_EN on Rev E pcb }                 ' extra port 0 - not on Rev E or lower
  XP1     = 24  { I/O }                                         ' extra port 1 - not on Rev E or lower 
  XP2     = 19  { I/O }                                         ' extra port 2 - not on Rev E or lower    
  XP3     = 16  { I/O }                                         ' extra port 3 - not on Rev E or lower 
                                                                 
  SIO_RX  = 26  { I/O }                                         ' TTL serial io
  SIO_TX  = 25  { I/O }                                         ' TTL serial io (not enabled)  
   
  RX2     = 23  { I }                                           ' RS-485 port
  RX2_EN  = 22  { O }                                                 
  TX2_EN  = 21  { O }                                                 
  TX2     = 20  { O }                                                 
                                                                 
  R_LED   = 18  { O }                                           ' bi-color LED
  G_LED   = 17  { O }                                                 
                                                               
  OUT7    = 15  { O }                                           ' outputs
  OUT6    = 14  { O }                                                 
  OUT5    = 13  { O }                                                 
  OUT4    = 12  { O }                                                 
  OUT3    = 11  { O }                                                 
  OUT2    = 10  { O }                                                 
  OUT1    =  9  { O }                                                 
  OUT0    =  8  { O }                                                 
                                                                 
  OPT_BR  =  7  { I/O }                                         ' options switches
  OPT_A1  =  6  { I/O }                                                  
  OPT_A0  =  5  { I/O }                                                  
  OPT_SM  =  4  { I/O }                                                  
                                                                 
  DMX_A8  =  3  { I }                                           ' high-bit of DMX address
                                                                 
  DO_165  =  2  { I }                                           ' control lines for 74x165
  CLK_165 =  1  { O }                                                  
  LD_165  =  0  { O }                                                  


con { app io }

  SDA2    = OPT_A0                                              ' secondary i2c (option port)
  SCL2    = OPT_SM
  

obj

' main                                                          ' * master Spin cog
  time       : "jm_time"                                        '   timing and delays
  rtimer     : "jm_time"                                        '   run timer for candles
  prng       : "jm_prng"                                        '   pseudo-random number generator
  io         : "jm_io_basic"                                    '   essential io
  rtc        : "jm_ds3231"                                      '   rtc
  pwm        : "jm_pwm8"                                        ' * pwm driver for outputs
  term       : "jm_fullduplexserial"                            ' * serial IO for terminal

' * uses cog when loaded

' flame generator cog runs during Menorah lighting -- unloaded when done
                                                             
                                                                  
var


dat

  ' This table holds the dates and sunset times for Hanukkah 2016-17
  ' -- source:  http://sunrise-sunset.org/us/burbank-ca

  Events        long    8                                       ' events in table

                '        yr mo dt   hr mn sc
  
                long    $16_12_24, $16_50_02, 1
                long    $16_12_25, $16_50_39, 2
                long    $16_12_26, $16_51_18, 3
                long    $16_12_27, $16_51_57, 4          
                long    $16_12_28, $16_52_38, 5
                long    $16_12_29, $16_53_21, 6
                long    $16_12_30, $16_54_04, 7
                long    $16_12_31, $16_54_49, 8
                  
  
pub main | last, dc, tc, day  
                                                                  
  setup                                                         ' setup io and objects
  
  set_clock_regs                                                ' check for clock update
  
  last := -1

  repeat                                                         
    repeat                                                      ' wait for new second
      dc := rtc.date_code
      tc := rtc.time_code
      if (tc <> last)
        last := tc
        quit

    term.tx(term#CLS)                                           ' show current time
    term.hex(tc.byte[2], 2)
    term.tx(":")
    term.hex(tc.byte[1], 2)  
    term.tx(":")             
    term.hex(tc.byte[0], 2)     

    day := scan_events(@Events, dc, tc)                         ' check events table
    if (day => 1)                                               ' if Hanukkah day
      term.tx(term#HOME)                                        '  show it and run Menorah
      term.str(string("Hanukkah Day "))
      term.tx("0" + day)

      light_menorah(day, 30)                                    ' set Menorah behavior here!


pub setup                                                  
                                                                  
'' Configure HC-8+ IO and support objects

  time.start                                                    ' setup for timing/delays
                                                                 
  prng.seed(cnt, cnt ~> 2, $EF8_7E4, cnt <- 2, cnt)             ' setup randomizer                             
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)
                                                                 
  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode
                                                                 
  if (PCB_REV =< "E")                                            
    io.low(OUT_EN)                                              ' enable output driver    

  rtc.start(SCL2, SDA2)                                         ' connect to DS3231
    
  pwm.start(8, OUT0)                                            ' start outputs driver *
                                                                 
  term.start(RX1, TX1, %0000, 115_200)                          ' start serial for terminal *


pub set_clock_regs | check, val

  time.pause(2000)                                              ' allow terminal to open

  term.tx(term#CLS)
  term.str(string("Current clock:"))
  term.str(string(13, "-- "))
  term.str(rtc.time_hhmmss(-1))
  term.str(string(13, "-- ")) 
  term.str(rtc.date_mmddyy(-1))
  term.str(string(13, "-- ")) 
  term.str(rtc.day_long(-1))
  repeat 2
    term.tx(13)

  term.str(string("Would you like to change? [y/N] :"))
  term.rxflush
  check := term.rxtime(3000)                                    ' wait 3s for response
  if ((check <> "y") and (check <> "Y"))
    return
  
  term.tx(term#CLS)
  term.str(string("Set clock registers", 13))

  term.str(string(13, "Hour (0..23)...... "))                   ' show register
  val := get_dec2(0, 23)                                        ' get value
  rtc.wr_reg(rtc#R_HRS, rtc.dec2bcd(val))                       ' set register in rtc

  term.str(string(13, "Minutes (0..59)... "))  
  val := get_dec2(0, 59)                            
  rtc.wr_reg(rtc#R_MNS, rtc.dec2bcd(val))
      
  term.str(string(13, "Seconds (0..59)... "))   
  val := get_dec2(0, 59)                               
  rtc.wr_reg(rtc#R_SCS, rtc.dec2bcd(val))   

  term.str(string(13, "Year (0..99)...... "))   
  val := get_dec2(0, 99)                               
  rtc.wr_reg(rtc#R_YR, rtc.dec2bcd(val))   

  term.str(string(13, "Month (1..12)..... "))   
  val := get_dec2(1, 12)                               
  rtc.wr_reg(rtc#R_MON, rtc.dec2bcd(val))   

  term.str(string(13, "Date (1..31)...... "))   
  val := get_dec2(1, 31)                               
  rtc.wr_reg(rtc#R_DATE, rtc.dec2bcd(val))   

  term.str(string(13, "Day (1..7)........ "))   
  val := get_dec2(1, 7)                               
  rtc.wr_reg(rtc#R_DAY, rtc.dec2bcd(val))

  rtc.wr_reg(rtc#R_CTRL,   $00)   
  rtc.wr_reg(rtc#R_STATUS, $00)


pub get_dec2(lo, hi) | value, len, c

  value := len := 0
  term.rxflush

  repeat
    c := term.rx
    case c
      "0".."9" :                                                ' digit           
        if (len < 2)                                                              
          value := (value * 10) + (c - "0")                                                         
          len += 1                                                                
          term.tx(c)                                                                  

      8 :                                                       ' backspace
        if (len > 0)
          value /= 10
          len -= 1
          term.tx(8)
          term.tx(" ")
          term.tx(8)
        
      13 :
        if ((value => lo) and (value =< hi))
          return value
   

pub scan_events(p_table, dc, tc) | count, tdc, ttc, tevent

'' Scans count event entries at p_table
'' -- if dc and tc match, returns event value (3rd entry)
'' -- assumes no more than one event per second (time code)

  count := long[p_table]                                        ' get events in table
  p_table += 4                                                  ' advance past count
  
  repeat count
    longmove(@tdc, p_table, 3)                                  ' grab table entry
    if (dc == tdc)                                              ' date code match?
      if (tc == ttc)                                            ' time code match?
        return tevent                                           ' return event code
    p_table += 12                                               ' skip 3 longs 

  return -1


var

  long  fgcog
  
  byte  flevel[8]


pub light_menorah(day, runtime) | ch, level, idx, tc, master

'' Light and run menorah candles for specified day (1..8)
'' if runtime > 0, run for that many minutes
'' -- if no run-time, run until midnight

  fgcog := cognew(flame_generator, @fgstack) + 1                ' start flame generator

  bytefill(@flevel, 0, 8)                                       ' zero flame levels
  
  ' sequence flames on
  
  repeat ch from 0 to day-1                                     ' run to current day
    repeat 255                                                  ' fade up to 100%
      time.set(-8)                                              ' 8ms * 255 = ~2s
      repeat
        refresh_candles(255)                     
      until (time.millis => 0)
      flevel[ch]++

    time.set(-3000)                                             ' 3-sec delay between candles
    repeat                   
      refresh_candles(255)        
    until (time.millis => 0) 

  if (runtime > 0)
    rtimer.start                                                ' start the run timer
    runtime *= 60                                               ' convert to seconds
    repeat                                                      ' update candles                                   
      refresh_candles(255) 
      if (rtimer.seconds => runtime)                            ' check timer
        quit                                                    ' quit when expired

  else { run until midnight }
    repeat                                       
      refresh_candles(255)              
      tc := rtc.time_code                                       ' get time 
      if (tc == $23_59_55)                                      ' approaching midnight?
        quit                                                    '  we're done

  ' fade out flames

  repeat master from 255 to 0                                   ' 100% to 0%
    time.set(-19)                                               ' ~5s fade out
    repeat
      refresh_candles(master)                  
    until (time.millis => 0)                                    ' quit when timer expired
      
  pwm.set_all(0)                                                ' ensure all off
  cogstop(fgcog - 1)                                            ' shut down flame generator
  fgcog := 0                      
    

pub refresh_candles(master) | idx, level

  repeat idx from 0 to 7                                        ' fresh all pwm channels               
    if (flevel[idx])                                            ' if this channel active                   
      level := flame[idx] * flevel[idx] * master / 65025        ' calculate level
    else                                 
      level := 0                         
    pwm.set(idx, level)                  


var

  long  fgstack[32]

  byte  flame[8]                                                ' flame level
  byte  target[8]
  

pri flame_generator | t, ch, mask                               ' call only using cognew()

  repeat ch from 0 to 7
    target[ch] := (prng.random >> 1) // 33 + 223                ' pre-seed high value

  repeat
    repeat ch from 0 to 7
      if (flame[ch] < target[ch])
        if (++flame[ch] == target[ch])
          target[ch] := (prng.random >> 1) // 33 + 159          ' low (159..191)

      else
        if (--flame[ch] == target[ch])
          target[ch] := (prng.random >> 1) // 33 + 223          ' high (223..255)       
          
    time.pause_us(3500)
     

dat { license }

{{

  Copyright (C) 2016 JonnyMac

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      